use crate::prelude::*;
use anyhow::Result;
use assetpath::AssetPath;

use cgmath::{Vector3, Zero};

use std::collections::HashMap;

use std::sync::Arc;

pub const ON_ENABLE_SOUND: &str = "OnCreate";
pub const ON_DISABLE_SOUND: &str = "OnDestroy";

pub struct Audio {
    custom_sounds: HashMap<String, Arc<Sound>>,

    current_state: Option<Arc<Sound>>,
    position: Vector3<f32>,
    context: Arc<Context>,
}

impl Audio {
    pub fn new(
        context: &Arc<Context>,
        sounds: impl Into<Option<HashMap<String, AssetPath>>>,
    ) -> Result<Self> {
        let mut sound_map = HashMap::new();

        if let Some(sounds) = sounds.into() {
            for (key, value) in sounds.into_iter() {
                if !value.is_empty() {
                    let sound =
                        context
                            .sound()
                            .load_sound(value, "sfx", SoundInterpretation::Spatial)?;
                    sound_map.insert(key, sound);
                }
            }
        }

        Ok(Self {
            custom_sounds: sound_map,

            current_state: None,
            position: Vector3::zero(),
            context: context.clone(),
        })
    }

    pub fn from_info(
        context: &Arc<Context>,
        data_prefix: &str,
        sounds: Vec<(String, AssetPath)>,
    ) -> Result<Self> {
        Self::new(
            context,
            sounds
                .into_iter()
                .map(|(name, mut path)| {
                    path.set_prefix(data_prefix);

                    (name, path)
                })
                .collect::<HashMap<String, AssetPath>>(),
        )
    }

    pub(crate) fn set_position(&mut self, position: Vector3<f32>) -> Result<()> {
        self.position = position;

        if let Some(audio) = &self.current_state {
            audio.set_position(position)?;
        }

        Ok(())
    }

    pub fn set_sound(&mut self, path: AssetPath, sound_type: &str) -> Result<()> {
        if let Some(audio) = self.custom_sounds.get_mut(sound_type) {
            *audio = self
                .context
                .sound()
                .load_sound(path, "sfx", SoundInterpretation::Spatial)?;
        }

        Ok(())
    }

    pub fn add_sound(&mut self, sound_type: &str, sound: Arc<Sound>) {
        self.custom_sounds.insert(sound_type.to_string(), sound);
    }

    pub fn play(&mut self, sound_type: &str, enable_looping: bool) -> Result<()> {
        if let Some(audio) = self.custom_sounds.get(sound_type) {
            audio.set_position(self.position)?;
            audio.play(Some(enable_looping))?;

            self.current_state = Some(audio.clone());
        }

        Ok(())
    }

    fn clear(&mut self) -> Result<()> {
        for sound in self.custom_sounds.values() {
            self.context.sound().remove_sound(sound)?;
        }

        Ok(())
    }

    pub fn sounds(&self) -> Vec<(String, AssetPath)> {
        let mut sounds = Vec::new();

        for (sound_type, sound) in self.custom_sounds.iter() {
            sounds.push((sound_type.clone(), sound.file_path().clone()));
        }

        sounds
    }

    pub fn audio_map(&self) -> Result<HashMap<String, AssetPath>> {
        let mut map = HashMap::new();

        for (sound_type, sound) in self.custom_sounds.iter() {
            map.insert(sound_type.clone(), sound.file_path().clone());
        }

        Ok(map)
    }

    pub fn stop_looping(&self) -> Result<()> {
        if let Some(sound) = &self.current_state {
            sound.stop_looping()?;
        }

        Ok(())
    }

    pub fn stop(&mut self) -> Result<()> {
        if let Some(sound) = &self.current_state {
            sound.stop()?;
        }

        self.current_state = None;

        Ok(())
    }

    pub fn add_custom_sound(&mut self, path: AssetPath, sound_type: &str) -> Result<()> {
        self.custom_sounds.insert(
            sound_type.to_string(),
            self.context
                .sound()
                .load_sound(path, "sfx", SoundInterpretation::Spatial)?,
        );

        Ok(())
    }
}

impl EntityComponent for Audio {
    fn name(&self) -> &str {
        Self::debug_name()
    }

    fn enable(&mut self, _world: &mut World) -> Result<()> {
        self.play(ON_ENABLE_SOUND, false)
    }

    fn disable(&mut self, _world: &mut World) -> Result<()> {
        self.play(ON_DISABLE_SOUND, false)
    }
}

impl ComponentDebug for Audio {
    fn debug_name() -> &'static str {
        "Audio"
    }
}

impl Drop for Audio {
    fn drop(&mut self) {
        if let Some(sound) = &self.current_state {
            sound.stop_looping().unwrap();
        }

        self.clear().unwrap();
    }
}
