use anyhow::Result;
use engine::prelude::*;
use std::slice::Iter;
use std::sync::Arc;

use crate::{
    config::save_game::SaveGame,
    items::{
        ability_addon::AbilityAddon,
        ability_book::{Ability, AbilityBook},
        Item, ItemSystem, Jewel, MapItem, Rarities,
    },
};

pub trait Storable {
    fn rarity(&self) -> Rarities;
    fn icon(&self) -> Arc<Image>;
}

#[derive(Default, Clone)]
pub struct Inventory<A: Ability> {
    items: Vec<Item>,
    addons: Vec<AbilityAddon>,
    books: Vec<AbilityBook<A>>,
    jewels: Vec<Jewel>,
    maps: Vec<MapItem>,
}

impl<A: Ability> Inventory<A> {
    // ------- items --------
    pub fn add_item(&mut self, item: Item) {
        self.items.push(item);
    }

    pub fn insert_item(&mut self, item: Item, index: usize) {
        self.items.insert(index, item);
    }

    pub fn remove_item(&mut self, index: usize) -> Item {
        self.items.remove(index)
    }

    pub fn iter_items(&self) -> Iter<'_, Item> {
        self.items.iter()
    }

    pub fn item_at(&self, index: usize) -> &Item {
        &self.items[index]
    }

    pub fn item_mut_at(&mut self, index: usize) -> &mut Item {
        &mut self.items[index]
    }

    // ------- jewels --------
    pub fn add_jewel(&mut self, jewel: Jewel) {
        self.jewels.push(jewel);
    }

    pub fn insert_jewel(&mut self, jewel: Jewel, index: usize) {
        self.jewels.insert(index, jewel);
    }

    pub fn remove_jewel(&mut self, index: usize) -> Jewel {
        self.jewels.remove(index)
    }

    pub fn iter_jewels(&self) -> Iter<'_, Jewel> {
        self.jewels.iter()
    }

    pub fn jewel_at(&self, index: usize) -> &Jewel {
        &self.jewels[index]
    }

    pub fn jewel_mut_at(&mut self, index: usize) -> &mut Jewel {
        &mut self.jewels[index]
    }

    // ------- maps --------
    pub fn add_map(&mut self, map: MapItem) {
        self.maps.push(map);
    }

    pub fn insert_map(&mut self, map: MapItem, index: usize) {
        self.maps.insert(index, map);
    }

    pub fn remove_map(&mut self, index: usize) -> MapItem {
        self.maps.remove(index)
    }

    pub fn iter_maps(&self) -> Iter<'_, MapItem> {
        self.maps.iter()
    }

    pub fn maps_at(&self, index: usize) -> &MapItem {
        &self.maps[index]
    }

    // ------- addons --------
    pub fn add_addon(&mut self, addon: AbilityAddon) {
        self.addons.push(addon);
    }

    // pub fn insert_addon(&mut self, addon: AbilityAddon, index: usize) {
    //     self.addons.insert(index, addon);
    // }

    pub fn remove_addon(&mut self, index: usize) -> AbilityAddon {
        self.addons.remove(index)
    }

    pub fn addon(&self, index: usize) -> &AbilityAddon {
        &self.addons[index]
    }

    pub fn iter_addons(&self) -> Iter<'_, AbilityAddon> {
        self.addons.iter()
    }

    pub fn addon_at(&self, index: usize) -> &AbilityAddon {
        &self.addons[index]
    }

    // ------- books --------
    pub fn add_book(&mut self, book: AbilityBook<A>) {
        self.books.push(book);
    }

    pub fn insert_book(&mut self, book: AbilityBook<A>, index: usize) {
        self.books.insert(index, book);
    }

    pub fn remove_book(&mut self, index: usize) -> AbilityBook<A> {
        self.books.remove(index)
    }

    pub fn iter_books(&self) -> Iter<'_, AbilityBook<A>> {
        self.books.iter()
    }

    pub fn book_at(&self, index: usize) -> &AbilityBook<A> {
        &self.books[index]
    }
}

impl<A: Ability> Inventory<A> {
    pub fn load(save_game: &SaveGame, item_system: &ItemSystem<A>) -> Result<Self> {
        let mut inventory: Inventory<A> = Default::default();

        for item_string in save_game.inventory.items.iter() {
            if !item_string.is_empty() {
                inventory.add_item(Item::from_persistent(item_string.split('|'), item_system)?);
            }
        }

        for addon_string in save_game.inventory.addons.iter() {
            if !addon_string.is_empty() {
                inventory.add_addon(AbilityAddon::from_persistent(
                    addon_string.split('|'),
                    item_system,
                )?);
            }
        }

        for book_string in save_game.inventory.books.iter() {
            if !book_string.is_empty() {
                inventory.add_book(AbilityBook::from_persistent(
                    book_string.split('|'),
                    item_system,
                )?);
            }
        }

        for jewel_string in save_game.inventory.jewels.iter() {
            if !jewel_string.is_empty() {
                let mut jewel = Jewel::from_persistent(&mut jewel_string.split('|'))?;

                jewel.icon =
                    Some(item_system.jewel_icon(jewel.rarity, jewel.level, jewel.attribute));
                jewel.update_stat(&item_system.item_settings);

                inventory.add_jewel(jewel);
            }
        }

        for map_string in save_game.inventory.maps.iter() {
            if !map_string.is_empty() {
                inventory.add_map(MapItem::from_persistent(
                    map_string.split('|'),
                    item_system,
                )?)
            }
        }

        Ok(inventory)
    }

    pub fn store(&self, save_game: &mut SaveGame) {
        for item in self.items.iter() {
            save_game.inventory.items.push(item.into_persistent());
        }

        for addon in self.addons.iter() {
            save_game.inventory.addons.push(addon.into_persistent());
        }

        for book in self.books.iter() {
            save_game.inventory.books.push(book.into_persistent());
        }

        for jewel in self.jewels.iter() {
            save_game.inventory.jewels.push(jewel.into_persistent());
        }

        for map in self.maps.iter() {
            save_game.inventory.maps.push(map.into_persistent());
        }
    }
}

impl<A: Ability + 'static> EntityComponent for Inventory<A> {
    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl<A: Ability> ComponentDebug for Inventory<A> {
    fn debug_name() -> &'static str {
        "Inventory"
    }
}
