use anyhow::Result;
use engine::prelude::*;

use std::fmt;

use super::{attributes::Attributes, statistics::*};
use crate::components::macros::AttributeAssociation;
use crate::{config::items::ItemSettings, items::Rarities};

macro_rules! apply {
    (
        $self:ident, $attrib:ident,
        ($level: ident, $rarity_multiplier:ident, $item_settings:ident),
        [$($elem:ident: $ty:ty $(,)? )+]
    ) => {
        paste::paste! {
            match &mut $self {
                $(
                    Self::$elem(v) => {
                        v.set($level as $ty * $rarity_multiplier as $ty * $item_settings.$attrib.[< $elem:snake >]);
                    }
                )+
            }
        }
    };
}

#[derive(Debug, PartialEq, Eq)]
pub enum StrengthStatisticTypes {
    AirResistance(AirResistance),
    FireResistance(FireResistance),
    WaterResistance(WaterResistance),
    PhysicalDamage(PhysicalDamage),
    Health(Health),
    HealthRegeneration(HealthRegeneration),
}

impl StrengthStatisticTypes {
    // total amount of types
    pub const TYPE_COUNT: u32 = 6;

    pub fn random() -> Self {
        Random::range(0, Self::TYPE_COUNT).into()
    }

    pub fn apply_for_jewel(
        mut self,
        level: u32,
        rarity: Rarities,
        item_settings: &ItemSettings,
    ) -> Self {
        let rarity_multiplier = item_settings.jewel_rarity_multiplier.from_rarity(rarity);

        apply!(
            self,
            per_strength_stats,
            (level, rarity_multiplier, item_settings),
            [
                AirResistance: u32,
                FireResistance: u32,
                WaterResistance: u32,
                PhysicalDamage: u32,
                Health: f32,
                HealthRegeneration: f32
            ]
        );

        self
    }
}

impl From<u32> for StrengthStatisticTypes {
    fn from(n: u32) -> Self {
        match n {
            0 => Self::AirResistance(AirResistance::default()),
            1 => Self::FireResistance(FireResistance::default()),
            2 => Self::WaterResistance(WaterResistance::default()),
            3 => Self::PhysicalDamage(PhysicalDamage::default()),
            4 => Self::Health(Health::default()),
            5 => Self::HealthRegeneration(HealthRegeneration::default()),

            _ => panic!("can only convert number below TYPE_COUNT"),
        }
    }
}

#[derive(Debug, PartialEq, Eq)]
pub enum AgilityStatisticTypes {
    Armor(Armor),
    CriticalHitChance(CriticalHitChance),
    CriticalHitDamage(CriticalHitDamage),
}

impl AgilityStatisticTypes {
    // total amount of types
    pub const TYPE_COUNT: u32 = 3;

    pub fn random() -> Self {
        Random::range(0, Self::TYPE_COUNT).into()
    }

    pub fn apply_for_jewel(
        mut self,
        level: u32,
        rarity: Rarities,
        item_settings: &ItemSettings,
    ) -> Self {
        let rarity_multiplier = item_settings.jewel_rarity_multiplier.from_rarity(rarity);

        apply!(
            self,
            per_agility_stats,
            (level, rarity_multiplier, item_settings),
            [
                Armor: u32,
                CriticalHitChance: f32,
                CriticalHitDamage: f32
            ]
        );

        self
    }
}

impl From<u32> for AgilityStatisticTypes {
    fn from(n: u32) -> Self {
        match n {
            0 => Self::Armor(Armor::default()),
            1 => Self::CriticalHitChance(CriticalHitChance::default()),
            2 => Self::CriticalHitDamage(CriticalHitDamage::default()),

            _ => panic!("can only convert number below TYPE_COUNT"),
        }
    }
}

#[derive(Debug, PartialEq, Eq)]
pub enum IntelligenceStatisticTypes {
    AirDamage(AirDamage),
    FireDamage(FireDamage),
    WaterDamage(WaterDamage),
    Mana(Mana),
    ManaRegeneration(ManaRegeneration),
}

impl IntelligenceStatisticTypes {
    // total amount of types
    pub const TYPE_COUNT: u32 = 5;

    pub fn random() -> Self {
        Random::range(0, Self::TYPE_COUNT).into()
    }

    pub fn apply_for_jewel(
        mut self,
        level: u32,
        rarity: Rarities,
        item_settings: &ItemSettings,
    ) -> Self {
        let rarity_multiplier = item_settings.jewel_rarity_multiplier.from_rarity(rarity);

        apply!(
            self,
            per_intelligence_stats,
            (level, rarity_multiplier, item_settings),
            [
                AirDamage: u32,
                FireDamage: u32,
                WaterDamage: u32,
                Mana: f32,
                ManaRegeneration: f32,
            ]
        );

        self
    }
}

impl From<u32> for IntelligenceStatisticTypes {
    fn from(n: u32) -> Self {
        match n {
            0 => Self::AirDamage(AirDamage::default()),
            1 => Self::FireDamage(FireDamage::default()),
            2 => Self::WaterDamage(WaterDamage::default()),
            3 => Self::Mana(Mana::default()),
            4 => Self::ManaRegeneration(ManaRegeneration::default()),

            _ => panic!("can only convert number below TYPE_COUNT"),
        }
    }
}

#[derive(Debug, PartialEq, Eq, Clone, Serialize, Deserialize)]
pub enum StatisticType {
    AirResistance(AirResistance),
    AirDamage(AirDamage),

    FireResistance(FireResistance),
    FireDamage(FireDamage),

    WaterResistance(WaterResistance),
    WaterDamage(WaterDamage),

    Armor(Armor),
    PhysicalDamage(PhysicalDamage),

    CriticalHitChance(CriticalHitChance),
    CriticalHitDamage(CriticalHitDamage),

    Health(Health),
    HealthRegeneration(HealthRegeneration),

    Mana(Mana),
    ManaRegeneration(ManaRegeneration),
}

impl StatisticType {
    // total amount of types
    pub const TYPE_COUNT: u32 = 14;

    pub fn apply_for_item(&mut self, attributes: &Attributes, item_settings: &ItemSettings) {
        match self {
            // air
            StatisticType::AirResistance(air_resistance) => air_resistance.set(
                air_resistance.attribute_level(attributes)
                    * item_settings.per_strength_stats.air_resistance,
            ),

            StatisticType::AirDamage(air_damage) => air_damage.set(
                air_damage.attribute_level(attributes)
                    * item_settings.per_intelligence_stats.air_damage,
            ),

            // fire
            StatisticType::FireResistance(fire_resistance) => fire_resistance.set(
                fire_resistance.attribute_level(attributes)
                    * item_settings.per_strength_stats.fire_resistance,
            ),
            StatisticType::FireDamage(fire_damage) => fire_damage.set(
                fire_damage.attribute_level(attributes)
                    * item_settings.per_intelligence_stats.fire_damage,
            ),

            // water
            StatisticType::WaterResistance(water_resistance) => water_resistance.set(
                water_resistance.attribute_level(attributes)
                    * item_settings.per_strength_stats.water_resistance,
            ),
            StatisticType::WaterDamage(water_damage) => water_damage.set(
                water_damage.attribute_level(attributes)
                    * item_settings.per_intelligence_stats.water_damage,
            ),

            // physical
            StatisticType::Armor(armor) => {
                armor.set(armor.attribute_level(attributes) * item_settings.per_agility_stats.armor)
            }
            StatisticType::PhysicalDamage(physical_damage) => physical_damage.set(
                physical_damage.attribute_level(attributes)
                    * item_settings.per_strength_stats.physical_damage,
            ),

            // crit
            StatisticType::CriticalHitChance(crit_chance) => crit_chance.set(
                crit_chance.attribute_level(attributes) as f32
                    * item_settings.per_agility_stats.critical_hit_chance,
            ),
            StatisticType::CriticalHitDamage(crit_damage) => crit_damage.set(
                crit_damage.attribute_level(attributes) as f32
                    * item_settings.per_agility_stats.critical_hit_damage,
            ),

            // health
            StatisticType::Health(health) => health.set(
                health.attribute_level(attributes) as f32 * item_settings.per_strength_stats.health,
            ),
            StatisticType::HealthRegeneration(health_regeneration) => health_regeneration.set(
                health_regeneration.attribute_level(attributes) as f32
                    * item_settings.per_strength_stats.health_regeneration,
            ),

            // mana
            StatisticType::Mana(mana) => mana.set(
                mana.attribute_level(attributes) as f32 * item_settings.per_intelligence_stats.mana,
            ),
            StatisticType::ManaRegeneration(mana_regeneration) => mana_regeneration.set(
                mana_regeneration.attribute_level(attributes) as f32
                    * item_settings.per_intelligence_stats.mana_regeneration,
            ),
        }
    }

    pub fn apply_for_jewel(&mut self, rarity: Rarities, level: u32, item_settings: &ItemSettings) {
        let factor = level * item_settings.jewel_rarity_multiplier.from_rarity(rarity);

        match self {
            // air
            StatisticType::AirResistance(air_resistance) => {
                air_resistance.set(factor * item_settings.per_strength_stats.air_resistance)
            }

            StatisticType::AirDamage(air_damage) => {
                air_damage.set(factor * item_settings.per_intelligence_stats.air_damage)
            }

            // fire
            StatisticType::FireResistance(fire_resistance) => {
                fire_resistance.set(factor * item_settings.per_strength_stats.fire_resistance)
            }
            StatisticType::FireDamage(fire_damage) => {
                fire_damage.set(factor * item_settings.per_intelligence_stats.fire_damage)
            }

            // water
            StatisticType::WaterResistance(water_resistance) => {
                water_resistance.set(factor * item_settings.per_strength_stats.water_resistance)
            }
            StatisticType::WaterDamage(water_damage) => {
                water_damage.set(factor * item_settings.per_intelligence_stats.water_damage)
            }

            // physical
            StatisticType::Armor(armor) => {
                armor.set(factor * item_settings.per_agility_stats.armor)
            }
            StatisticType::PhysicalDamage(physical_damage) => {
                physical_damage.set(factor * item_settings.per_strength_stats.physical_damage)
            }

            // crit
            StatisticType::CriticalHitChance(crit_chance) => {
                crit_chance.set(factor as f32 * item_settings.per_agility_stats.critical_hit_chance)
            }
            StatisticType::CriticalHitDamage(crit_damage) => {
                crit_damage.set(factor as f32 * item_settings.per_agility_stats.critical_hit_damage)
            }

            // health
            StatisticType::Health(health) => {
                health.set(factor as f32 * item_settings.per_strength_stats.health)
            }
            StatisticType::HealthRegeneration(health_regeneration) => health_regeneration
                .set(factor as f32 * item_settings.per_strength_stats.health_regeneration),

            // mana
            StatisticType::Mana(mana) => {
                mana.set(factor as f32 * item_settings.per_intelligence_stats.mana)
            }
            StatisticType::ManaRegeneration(mana_regeneration) => mana_regeneration
                .set(factor as f32 * item_settings.per_intelligence_stats.mana_regeneration),
        }
    }

    pub fn same_type(&self, other: &Self) -> bool {
        matches!(
            (self, other),
            (Self::AirResistance(_), Self::AirResistance(_))
                | (Self::AirDamage(_), Self::AirDamage(_))
                | (Self::FireResistance(_), Self::FireResistance(_))
                | (Self::FireDamage(_), Self::FireDamage(_))
                | (Self::WaterResistance(_), Self::WaterResistance(_))
                | (Self::WaterDamage(_), Self::WaterDamage(_))
                | (Self::Armor(_), Self::Armor(_))
                | (Self::PhysicalDamage(_), Self::PhysicalDamage(_))
                | (Self::CriticalHitChance(_), Self::CriticalHitChance(_))
                | (Self::CriticalHitDamage(_), Self::CriticalHitDamage(_))
                | (Self::Health(_), Self::Health(_))
                | (Self::HealthRegeneration(_), Self::HealthRegeneration(_))
                | (Self::Mana(_), Self::Mana(_))
                | (Self::ManaRegeneration(_), Self::ManaRegeneration(_))
        )
    }

    pub fn display_value(&self) -> String {
        match *self {
            // air
            StatisticType::AirResistance(v) => format!("{}", v.raw()),
            StatisticType::AirDamage(v) => format!("{}", v.raw()),

            // fire
            StatisticType::FireResistance(v) => format!("{}", v.raw()),
            StatisticType::FireDamage(v) => format!("{}", v.raw()),

            // water
            StatisticType::WaterResistance(v) => format!("{}", v.raw()),
            StatisticType::WaterDamage(v) => format!("{}", v.raw()),

            // physical
            StatisticType::Armor(v) => format!("{}", v.raw()),
            StatisticType::PhysicalDamage(v) => format!("{}", v.raw()),

            // crit
            StatisticType::CriticalHitChance(v) => format!("{:.2} %", v.raw()),
            StatisticType::CriticalHitDamage(v) => format!("{:.2} %", v.raw()),

            // health
            StatisticType::Health(v) => format!("{:.0}", v.raw()),
            StatisticType::HealthRegeneration(v) => format!("{:.2}", v.raw()),

            // mana
            StatisticType::Mana(v) => format!("{:.0}", v.raw()),
            StatisticType::ManaRegeneration(v) => format!("{:.2}", v.raw()),
        }
    }
}

impl std::ops::AddAssign for StatisticType {
    fn add_assign(&mut self, other: StatisticType) {
        match (self, &other) {
            (Self::AirResistance(me), Self::AirResistance(val)) => *me += *val,
            (Self::AirDamage(me), Self::AirDamage(val)) => *me += *val,

            (Self::FireResistance(me), Self::FireResistance(val)) => *me += *val,
            (Self::FireDamage(me), Self::FireDamage(val)) => *me += *val,

            (Self::WaterResistance(me), Self::WaterResistance(val)) => *me += *val,
            (Self::WaterDamage(me), Self::WaterDamage(val)) => *me += *val,

            (Self::Armor(me), Self::Armor(val)) => *me += *val,
            (Self::PhysicalDamage(me), Self::PhysicalDamage(val)) => *me += *val,

            (Self::CriticalHitChance(me), Self::CriticalHitChance(val)) => *me += *val,
            (Self::CriticalHitDamage(me), Self::CriticalHitDamage(val)) => *me += *val,

            (Self::Health(me), Self::Health(val)) => *me += *val,
            (Self::HealthRegeneration(me), Self::HealthRegeneration(val)) => *me += *val,

            (Self::Mana(me), Self::Mana(val)) => *me += *val,
            (Self::ManaRegeneration(me), Self::ManaRegeneration(val)) => *me += *val,

            _ => panic!("can not add different types"),
        }
    }
}

impl From<u32> for StatisticType {
    fn from(n: u32) -> Self {
        match n {
            // air
            0 => Self::AirResistance(AirResistance::default()),
            1 => Self::AirDamage(AirDamage::default()),

            // fire
            2 => Self::FireResistance(FireResistance::default()),
            3 => Self::FireDamage(FireDamage::default()),

            // water
            4 => Self::WaterResistance(WaterResistance::default()),
            5 => Self::WaterDamage(WaterDamage::default()),

            // physical
            6 => Self::Armor(Armor::default()),
            7 => Self::PhysicalDamage(PhysicalDamage::default()),

            // Crit
            8 => Self::CriticalHitChance(CriticalHitChance::default()),
            9 => Self::CriticalHitDamage(CriticalHitDamage::default()),

            // health
            10 => Self::Health(Health::default()),
            11 => Self::HealthRegeneration(HealthRegeneration::default()),

            // mana
            12 => Self::Mana(Mana::default()),
            13 => Self::ManaRegeneration(ManaRegeneration::default()),

            _ => panic!("can only convert number below TYPE_COUNT"),
        }
    }
}

impl From<AgilityStatisticTypes> for StatisticType {
    fn from(agility_statistics: AgilityStatisticTypes) -> Self {
        match agility_statistics {
            AgilityStatisticTypes::CriticalHitChance(v) => StatisticType::CriticalHitChance(v),
            AgilityStatisticTypes::CriticalHitDamage(v) => StatisticType::CriticalHitDamage(v),
            AgilityStatisticTypes::Armor(v) => StatisticType::Armor(v),
        }
    }
}

impl From<IntelligenceStatisticTypes> for StatisticType {
    fn from(intelligence_statistics: IntelligenceStatisticTypes) -> Self {
        match intelligence_statistics {
            IntelligenceStatisticTypes::AirDamage(v) => StatisticType::AirDamage(v),
            IntelligenceStatisticTypes::FireDamage(v) => StatisticType::FireDamage(v),
            IntelligenceStatisticTypes::WaterDamage(v) => StatisticType::WaterDamage(v),
            IntelligenceStatisticTypes::Mana(v) => StatisticType::Mana(v),
            IntelligenceStatisticTypes::ManaRegeneration(v) => StatisticType::ManaRegeneration(v),
        }
    }
}

impl From<StrengthStatisticTypes> for StatisticType {
    fn from(strength_statistics: StrengthStatisticTypes) -> Self {
        match strength_statistics {
            StrengthStatisticTypes::AirResistance(v) => StatisticType::AirResistance(v),
            StrengthStatisticTypes::FireResistance(v) => StatisticType::FireResistance(v),
            StrengthStatisticTypes::WaterResistance(v) => StatisticType::WaterResistance(v),
            StrengthStatisticTypes::PhysicalDamage(v) => StatisticType::PhysicalDamage(v),
            StrengthStatisticTypes::Health(v) => StatisticType::Health(v),
            StrengthStatisticTypes::HealthRegeneration(v) => StatisticType::HealthRegeneration(v),
        }
    }
}

impl std::str::FromStr for StatisticType {
    type Err = anyhow::Error;

    fn from_str(s: &str) -> Result<Self> {
        match s {
            "Air Resistance" => Ok(Self::AirResistance(AirResistance::default())),
            "Air Damage" => Ok(Self::AirDamage(AirDamage::default())),

            // fire
            "Fire Resistance" => Ok(Self::FireResistance(FireResistance::default())),
            "Fire Damage" => Ok(Self::FireDamage(FireDamage::default())),

            // water
            "Water Resistance" => Ok(Self::WaterResistance(WaterResistance::default())),
            "Water Damage" => Ok(Self::WaterDamage(WaterDamage::default())),

            // physical
            "Armor" => Ok(Self::Armor(Armor::default())),
            "Physical Damage" => Ok(Self::PhysicalDamage(PhysicalDamage::default())),

            // Crit
            "Critical Hit Chance" => Ok(Self::CriticalHitChance(CriticalHitChance::default())),
            "Critical Hit Damage" => Ok(Self::CriticalHitDamage(CriticalHitDamage::default())),

            // health
            "Health" => Ok(Self::Health(Health::default())),
            "Health Regeneration" => Ok(Self::HealthRegeneration(HealthRegeneration::default())),

            // mana
            "Mana" => Ok(Self::Mana(Mana::default())),
            "Mana Regeneration" => Ok(Self::ManaRegeneration(ManaRegeneration::default())),

            _ => {
                return Err(anyhow::Error::msg(format!(
                    "Failed parsing StatisticTypes from {}",
                    s
                )))
            }
        }
    }
}

impl fmt::Display for StatisticType {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match *self {
            // air
            StatisticType::AirResistance(_) => write!(f, "Air Resistance"),
            StatisticType::AirDamage(_) => write!(f, "Air Damage"),

            // fire
            StatisticType::FireResistance(_) => write!(f, "Fire Resistance"),
            StatisticType::FireDamage(_) => write!(f, "Fire Damage"),

            // water
            StatisticType::WaterResistance(_) => write!(f, "Water Resistance"),
            StatisticType::WaterDamage(_) => write!(f, "Water Damage"),

            // physical
            StatisticType::Armor(_) => write!(f, "Armor"),
            StatisticType::PhysicalDamage(_) => write!(f, "Physical Damage"),

            // crit
            StatisticType::CriticalHitChance(_) => write!(f, "Critical Hit Chance"),
            StatisticType::CriticalHitDamage(_) => write!(f, "Critical Hit Damage"),

            // health
            StatisticType::Health(_) => write!(f, "Health"),
            StatisticType::HealthRegeneration(_) => write!(f, "Health Regeneration"),

            // mana
            StatisticType::Mana(_) => write!(f, "Mana"),
            StatisticType::ManaRegeneration(_) => write!(f, "Mana Regeneration"),
        }
    }
}
